! MODULES AND SUBROUTINES FOR LENSING NFW PROFILE FITTING

module constants
  implicit none
  DOUBLE PRECISION, PARAMETER :: ckms = 3.0d5 ! speed of light in km/s
  DOUBLE PRECISION, PARAMETER :: rhocrit0=2.7751973751261264d11 ! rhocrit in units of h^-1 Msun/ h^-3 Mpc^3
  DOUBLE PRECISION, PARAMETER :: Gs = 4.3d-9 ! G in units of Mpc/Msun /(km/s)^2
  DOUBLE PRECISION, PARAMETER :: pi=4.0d0*datan(1.0d0)

! Once the redshift of the lens is known these numbers are constants throughtout the entire code
  DOUBLE PRECISION, SAVE :: dAlens,rhoclens

end module constants
! ####################################################################
module cosmology
  USE CONSTANTS
  implicit none


! THE COSMOLOGICAL PARAMETERS
! TRIVIAL FOR A FLAT LCDM UNIVERSE WITH LENGTH UNITS OF h^-1 Mpc

  public
  TYPE cospar
     DOUBLE PRECISION :: omegam
  END TYPE cospar
  Type (cospar), SAVE :: cosmopar
  
contains


!-----------------------------------------------------------------------

! E=H(z)/H0
  function Eh(z)
    DOUBLE PRECISION :: Eh
    DOUBLE PRECISION, intent(in) :: z
    DOUBLE PRECISION ::  a
    a = 1.0d0/(1.0d0+z)
    Eh=dsqrt(cosmopar%omegam/a/a/a+(1.0d0-cosmopar%omegam))
    RETURN
  end function Eh


!-----------------------------------------------------------------------


! A USEFUL ROMBERG INTEGRATION ROUTINE
!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
        function rombint(f,a,b,tol)
!  Rombint returns the integral from a to b of using Romberg integration.
!  The method converges provided that f(x) is continuous in (a,b).
!  f must be DOUBLE PRECISION and must be declared external in the calling
!  routine.  tol indicates the desired relative accuracy in the integral.
!
        implicit none
        integer, parameter :: MAXITER=20,MAXJ=5
        dimension g(MAXJ+1)
        DOUBLE PRECISION f
        external f
        DOUBLE PRECISION :: rombint
        DOUBLE PRECISION, intent(in) :: a,b,tol
        integer :: nint, i, k, jmax, j
        DOUBLE PRECISION :: h, gmax, error, g, g0, g1, fourj
!
        h=0.5d0*(b-a)
        gmax=h*(f(a)+f(b))
        g(1)=gmax
        nint=1
        error=1.0d20
        i=0
10        i=i+1
          if (i.gt.MAXITER.or.(i.gt.5.and.abs(error).lt.tol)) &
            go to 40
!  Calculate next trapezoidal rule approximation to integral.

          g0=0.0
            do 20 k=1,nint
            g0=g0+f(a+(k+k-1)*h)
20        continue
          g0=0.5d0*g(1)+h*g0
          h=0.5d0*h
          nint=nint+nint
          jmax=min(i,MAXJ)
          fourj=1.0
            do 30 j=1,jmax
!  Use Richardson extrapolation.
            fourj=4.0*fourj
            g1=g0+(g0-g(j))/(fourj-1.0)
            g(j)=g0
            g0=g1
30        continue
          if (abs(g0).gt.tol) then
            error=1.0-gmax/g0
          else
            error=gmax
          end if
          gmax=g0
          g(jmax+1)=g0
        go to 10
40      rombint=g0
        if (i.gt.MAXITER.and.abs(error).gt.tol)  then
          write(*,*) 'Warning: Rombint failed to converge; '
          write (*,*)'integral, error, tol:', rombint,error, tol
        end if
        
        end function rombint


!-----------------------------------------------------------------------

! coordinate distance , in units of h^-1 Mpc

  function r(z)
    DOUBLE PRECISION, INTENT(IN) :: z
    DOUBLE PRECISION :: r
    DOUBLE PRECISION, PARAMETER :: tol=1.0d-8
    DOUBLE PRECISION :: integral

    integral=rombint(rint,0.0d0,z,tol)
    r= ckms*integral/1.0d2
    
    return
  end function r

!-----------------------------------------------------------------------

! integrand for coordinate distance 
  function rint(z)
    DOUBLE PRECISION, INTENT(IN) :: z
    DOUBLE PRECISION :: rint
    rint = 1.0d0/Eh(z)
    RETURN
  end function rint


!-----------------------------------------------------------------------


! angular diameter distance in units of h^-1 Mpc
  function dA(z)
    DOUBLE PRECISION, INTENT(IN) :: z
    DOUBLE PRECISION :: dA
    dA = r(z)/(1.+z)
    RETURN
  end function dA
!-----------------------------------------------------------------------    


! critical density as a function of redshift
  function rhocrit(z)
    DOUBLE PRECISION :: rhocrit
    DOUBLE PRECISION, INTENT(IN) :: z
    DOUBLE PRECISION :: hdim
    
    hdim = Eh(z)
    rhocrit = rhocrit0*hdim*hdim

    RETURN
  end function rhocrit

end module cosmology

! ####################################################################


! MODULE TO CALCULATE PROFILE QUANTITIES
module profile
  USE CONSTANTS
  USE COSMOLOGY
  IMPLICIT NONE
  

  contains


! Critical Surface Density in units of hMsun/Mpc^2
!  zd: redshift of lense 
!  zs: measured redshift of objekt
!  returns: Critical Surface Density (Sigma_c)
    function Sigmac(zd,zs)
      DOUBLE PRECISION :: Sigmac
      DOUBLE PRECISION, INTENT(IN) :: zd,zs
      DOUBLE PRECISION :: Ds,Dd,Dds

      Ds=dA(zs)
      Dd=dAlens

! The following only works for flat Universe
      Dds = Ds-(1.0d0+zd)*Dd/(1.0d0+zs)
      
      Sigmac = 0.25d0*ckms*ckms*Ds/Dd/Dds/pi/Gs
      
      RETURN
    end function Sigmac

!-----------------------------------------------------------------------        

! calculate density contrast of NFW Profile
!  conc:     concentration parameter
!  returns:  density contrast (delta_c)
    function deltac(conc)

      DOUBLE PRECISION :: deltac
      DOUBLE PRECISION, INTENT(IN) :: conc
   
      deltac = 200.0d0*conc*conc*conc/(dlog(1.0+conc)-conc/(1.0d0+conc))/3.0d0

      RETURN
    end function deltac

!-----------------------------------------------------------------------    


! G function in the shear calculation
!  x:  normalised radial distance (x=r/r_s)
    function ggam(x)
      DOUBLE PRECISION :: ggam
      DOUBLE PRECISION, INTENT(IN) :: x
      DOUBLE PRECISION :: ath,arct
      IF (x<1.0d0) then
         ath=datanh(dsqrt((1.0d0-x)/(1.0d0+x)))
         ggam=8.0d0*ath/x/x/dsqrt(1.0d0-x*x)+4.0d0/x/x*dlog(x/2.0d0)-2.0d0/(x*x-1.0d0)+4.0d0*ath/(x*x-1.0d0)/dsqrt(1.0d0-x*x)
      elseif (x>1.0d0) then
         arct=datan(dsqrt((x-1.0d0)/(1.0d0+x)))
         ggam=8.0d0*arct/x/x/dsqrt(x*x-1.0d0)+4.0d0/x/x*dlog(x/2.0d0)-2.0d0/(x*x-1.0d0)+4.0d0*arct/(x*x-1.0d0)**1.5d0
      else
         ggam=10.0d0/3.0d0-4.0d0*dlog(2.0d0)
      end IF


      RETURN
    end function ggam
!-----------------------------------------------------------------------    

! Tangential gravitational shearing
!  x:  normalised radial distance (x=r/r_s)
!  zd: redshift of lense 
!  zs: measured redshift of objekt
!  rs: scale radius (r_s)
!  dc: density contrast (delta_c)
!  returns: gravitational shear (gamma)
    function gshear(x,zd,zs,rs,dc)
      DOUBLE PRECISION :: gshear
      DOUBLE PRECISION, INTENT(IN) :: x,zd,zs,rs,dc
            
      gshear=rs*dc*rhoclens*ggam(x)/Sigmac(zd,zs)
      
      RETURN
    end function gshear

end module profile


!-----------------------------------------------------------------------    

! MODULE TO READ DATA FILE AND CALCULATE INPUT QUANTITIES
! Saves the data into following arrays:
!  ZDAT:      redshift of object
!  SHEARDAT:  measured shear of object
!  THETADAT:  angel at with the object is seen
!  RDAT:      distance of the object from the viewing axis

MODULE INPUTDATA
  USE COSMOLOGY
  USE CONSTANTS
  USE PROFILE
  
  IMPLICIT NONE
  DOUBLE PRECISION, ALLOCATABLE,SAVE :: ZDAT(:),SHEARDAT(:),THETADAT(:),RDAT(:)
  DOUBLE PRECISION, SAVE :: sigmashear
  DOUBLE PRECISION, SAVE :: zlens

  INTEGER, SAVE :: NDAT

  CONTAINS

!-----------------------------------------------------------------------    

    SUBROUTINE READINPUT(FILENAME)
      CHARACTER*80 :: FILENAME
      INTEGER :: I
      DOUBLE PRECISION :: DDUMMY
      INTEGER :: IDUMMY
      CHARACTER*80 :: DUMMYSTRING
      OPEN(UNIT=12,FILE=FILENAME,STATUS='unknown',FORM='formatted')
      READ(12,*) DUMMYSTRING
      READ(12,*) DUMMYSTRING
      IF (ALLOCATED(ZDAT)) DEALLOCATE(ZDAT)
      IF (ALLOCATED(SHEARDAT)) DEALLOCATE(SHEARDAT)
      IF (ALLOCATED(THETADAT)) DEALLOCATE(THETADAT)
      IF (ALLOCATED(RDAT)) DEALLOCATE(RDAT)
      ALLOCATE(THETADAT(NDAT),ZDAT(NDAT),SHEARDAT(NDAT),RDAT(NDAT))

      DO I=1,NDAT
         READ(12,*) IDUMMY,DDUMMY,DDUMMY,THETADAT(i),DDUMMY,SHEARDAT(i),DDUMMY,DDUMMY,DDUMMY,DDUMMY,DDUMMY,DDUMMY,DDUMMY,ZDAT(I)
      END DO
      CLOSE(12)

!!DIR$ PARALLEL
!!dir$ loop count min(256)
      RDAT=dA(zlens)*THETADAT*pi/180.0d0/60.0d0

      RETURN
      
    END SUBROUTINE READINPUT

!-----------------------------------------------------------------------    

! Calulate scale radius (r_s) of an NFW-profile with given mass (M_200)
! and concentration parameter (c).
!  M200: virial mass inside of r_200 (M_200)
!  con:  concentration parameter (c)
!  returns: scale radius (r_s)
    function rsnfw(M200,con)
      DOUBLE PRECISION, INTENT(IN) :: M200,con
      DOUBLE PRECISION :: rsnfw
      
      rsnfw = (3.0d0*M200/800.0d0/pi/rhoclens/con/con/con)**(1.0d0/3.0d0)

    end function rsnfw
    

END MODULE INPUTDATA

!-----------------------------------------------------------------------    

! MODULE FOR THE CHI2 CALCULATION
MODULE CHI2CALC

  USE COSMOLOGY
  USE INPUTDATA
  IMPLICIT NONE

contains


!-----------------------------------------------------------------------    
! calculate chi-squared for a given parameter set
!  rs:  scale radius (r_s)
!  dc:  density contrast (delta_c)
!  returns: chi-squared (chi^2) for the given parameter set and loaded data
  FUNCTION CHI2(rs,dc)
    DOUBLE PRECISION, INTENT(IN) :: rs,dc
    DOUBLE PRECISION :: CHI2
    INTEGER :: I
    DOUBLE PRECISION :: sheartk
    DOUBLE PRECISION :: x

    CHI2=1
    !you must edit this file. Rember to call 
    do I=1, NDAT

      write(*,*) 'help statment!',I,RDAT(i),ZDAT(i)
      CHI2=CHI2+1
    enddo
    RETURN
  END FUNCTION CHI2


END MODULE CHI2CALC

 !-----------------------------------------------------------------------     

! THE MARKOV CHAIN MODULE

MODULE MCMC

  INTEGER, SAVE :: NPAR
  INTEGER, SAVE :: NSAMPLE
  DOUBLE PRECISION, ALLOCATABLE :: PAR(:)
  DOUBLE PRECISION, ALLOCATABLE :: DPAR(:),LOBOUND(:),UPBOUND(:),DPARI(:)

CONTAINS


! GENERATION OF GAUSSIAN RANDOM NUMBER, USES THE MACHINE ROUTINE RANDOM_NUMBER
! FOR GENERATING UNIFORM DIST. RANDOM NUMBERS
!  A: array wich should be filled with random numbers
  SUBROUTINE GAUSSRANDOM(A)
    DOUBLE PRECISION :: A(:)
    INTEGER :: NDIM,I
    DOUBLE PRECISION :: v1,v2,rsq,FAC
    REAL :: R1,R2
    NDIM=SIZE(A)
    DO I=1,NDIM
       rsq=1.5
       DO WHILE ((rsq>=1.0).OR.(rsq==0.0))  
          CALL RANDOM_NUMBER(R1)
          CALL RANDOM_NUMBER(R2)
          v1 = 2.0*DBLE(R1)-1.0d0
          v2 = 2.0*DBLE(R2)-1.0d0
          rsq = v1**2+v2**2
       END DO
       FAC = DSQRT(-2.0d0*DLOG(rsq)/rsq)
       A(I) = v1*fac
    END DO
       
  END SUBROUTINE GAUSSRANDOM

!-----------------------------------------------------------------------    
! THE ACTUAL CHAIN ROUTINE, SAVED TO FILENAME

! INITIALIZE RANDOM_SEED BASED ON SYSTEM TIME
  SUBROUTINE init_random_seed()
    INTEGER :: i, n, clock
    INTEGER, DIMENSION(:), ALLOCATABLE :: seed
    
    CALL RANDOM_SEED(size = n)
    ALLOCATE(seed(n))
          
    CALL SYSTEM_CLOCK(COUNT=clock)
    
    seed = clock + 37 * (/ (i - 1, i = 1, n) /)
    CALL RANDOM_SEED(PUT = seed)
          
    DEALLOCATE(seed)
  END SUBROUTINE init_random_seed


  SUBROUTINE CHAIN(FILENAME)
    
    INTEGER :: I,NP,COUNT
    DOUBLE PRECISION :: CHI2GET,CHIOLD
    DOUBLE PRECISION, ALLOCATABLE :: PAROLD(:),RSHIFT(:)
    DOUBLE PRECISION :: ACCEPT,ATEST
    CHARACTER(80) :: FILENAME


    ALLOCATE(PAROLD(NPAR),RSHIFT(NPAR))
!    CALL RANDOM_SEED
    CALL init_random_seed()





    DEALLOCATE(PAROLD,RSHIFT)
    
    RETURN
  END SUBROUTINE CHAIN

!-----------------------------------------------------------------------    


! ASSOCIATE PHYSICAL WITH GENERIC PARAMETERS
  SUBROUTINE MATCHCHI2(PAR,CHI2PASS)
    USE CONSTANTS
    USE CHI2CALC
    USE PROFILE
    DOUBLE PRECISION :: PAR(:)
    DOUBLE PRECISION :: CHI2PASS
    DOUBLE PRECISION :: conc,rs,dc,M200
    M200 = PAR(1)
    conc = PAR(2)
    rs=rsnfw(M200,conc)
    dc=deltac(conc)
    CHI2PASS = CHI2(rs,dc)
  END SUBROUTINE MATCHCHI2

!-----------------------------------------------------------------------    

! CHECK PARAMETER BOUNDS     
  SUBROUTINE CHECK(PAR)
    DOUBLE PRECISION :: PAR(:)
    DO i=1,NPAR
       IF (PAR(I).lt.LOBOUND(I)) THEN 
          PAR(I) = LOBOUND(I) 
       ELSEIF (PAR(I).gt.UPBOUND(I)) THEN
          PAR(I) = UPBOUND(I)
       END IF
    END DO
  END SUBROUTINE CHECK
       


END MODULE MCMC
